#lang racket

(provide (all-defined-out))

;; Zopakovanie.

(define (iter f n x)
  (if (= n 0) x (iter f (- n 1) (f x))))

(define (natural-number? x)
  (exact-nonnegative-integer? x))

;;; Univerzálna funkcia pre primitívne rekurzívne funkcie.

;;; Aritmetizácia primitívne rekurzívnych funkčných symbolov.

;; Konštruktory.

(define Zx (list 'Zx))                     ; konštantná funkcia λx.0
(define Sx (list 'Sx))                     ; funkcia nasledovníka λx.x+1
(define (Ix n i) (list 'Ix n i))           ; identity (projekcie)
(define (Px g gs) (list 'Px g gs))         ; kontrakcia
(define (Cx n m h gs) (list 'Cx n m h gs)) ; kompozícia
(define (Rx n g h) (list 'Rx n g h))       ; primitívna rekurzia
(define (Xc2 n m) (list 'Xc2 n m))         ; konštantné funkcie
(define Xpair (list 'Xpair))               ; párovacia funkcia

;; Predikáty.

(define (Zx? e) (equal? (list-ref e 0) 'Zx))
(define (Sx? e) (equal? (list-ref e 0) 'Sx))
(define (Ix? e) (equal? (list-ref e 0) 'Ix))
(define (Px? e) (equal? (list-ref e 0) 'Px))
(define (Cx? e) (equal? (list-ref e 0) 'Cx))
(define (Rx? e) (equal? (list-ref e 0) 'Rx))
(define (Xc2? e) (equal? (list-ref e 0) 'Xc2))
(define (Xpair? e) (equal? (list-ref e 0) 'Xpair))

;; Selektory.

(define (Ix-arity e) (list-ref e 1))
(define (Ix-i e) (list-ref e 2))

(define (Px-g e) (list-ref e 1))
(define (Px-gs e) (list-ref e 2))

(define (Cx-arity e) (list-ref e 1))
(define (Cx-h-arity e) (list-ref e 2))
(define (Cx-h e) (list-ref e 3))
(define (Cx-gs e) (list-ref e 4))

(define (Rx-arity e) (list-ref e 1))
(define (Rx-g e) (list-ref e 2))
(define (Rx-h e) (list-ref e 3))

(define (Xc2-arity e) (list-ref e 1))
(define (Xc2-m e) (list-ref e 2))

;; Údajový typ pre primitívne rekurzívne funkčné symboly.

(define (PRFS? e)
  (and (list? e)
       (>= (length e) 1)
       (or (and (Zx? e)
                (= (length e) 1))
           (and (Sx? e)
                (= (length e) 1))
           (and (Ix? e)
                (= (length e) 3)
                (natural-number? (Ix-arity e))
                (natural-number? (Ix-i e)))
           (and (Px? e)
                (= (length e) 3)
                (PRFS? (Px-g e))
                (PRFS? (Px-gs e)))
           (and (Cx? e)
                (= (length e) 5)
                (natural-number? (Cx-arity e))
                (natural-number? (Cx-h-arity e))
                (PRFS? (Cx-h e))
                (PRFS? (Cx-gs e)))
           (and (Rx? e)
                (= (length e) 4)
                (natural-number? (Cx-arity e))
                (PRFS? (Rx-g e))
                (PRFS? (Rx-h e)))
           (and (Xc2? e)
                (= (length e) 3)
                (natural-number? (Xc2-arity e))
                (natural-number? (Xc2-m e)))
           (and (Xpair? e)
                (= (length e) 1)))))

;; Predikát p.r.-function-symbol?(n,e) platí, ak číslo e je kódom n-árneho
;; primitívne rekurzívneho funkčného symbolu.

(define (p.r.-function-symbol? n e)
  (define (is-p.r.-function-symbol? n e)
    (or (and (Zx? e)
             (= n 1))
        (and (Sx? e)
             (= n 1))
        (and (Ix? e)
             (= (Ix-arity e) n)
             (<= 1 (Ix-i e))
             (<= (Ix-i e) n))
        (and (Cx? e)
             (= (Cx-arity e) n)
             (>= (Cx-h-arity e) 1)
             (is-p.r.-function-symbol? (Cx-h-arity e) (Cx-h e))
             (are-p.r.-function-symbols? n (Cx-gs e) (Cx-h-arity e)))
        (and (Rx? e)
             (= (Rx-arity e) n)
             (>= n 2)
             (is-p.r.-function-symbol? (- n 1) (Rx-g e))
             (is-p.r.-function-symbol? (+ n 1) (Rx-h e)))
        (and (Xc2? e)
             (= (Xc2-arity e) n))
        (and (Xpair? e)
             (= n 2))))
  (define (are-p.r.-function-symbols? n e k)
    (or (and (= k 1)
             (is-p.r.-function-symbol? n e))
        (and (> k 1)
             (Px? e)
             (is-p.r.-function-symbol? n (Px-g e))
             (are-p.r.-function-symbols? n (Px-gs e) (- k 1)))))
  (and (PRFS? e) (>= n 1) (is-p.r.-function-symbol? n e)))

;; Arita primitívne rekurzívnych funkčných symbolov.

(define (arity e)
  (cond [(Zx? e) 1]
        [(Sx? e) 1]
        [(Ix? e) (Ix-arity e)]
        [(Cx? e) (Cx-arity e)]
        [(Rx? e) (Rx-arity e)]
        [(Xc2? e) Xc2-arity e]
        [(Xpair? e) 2]))

;;; Interpreter primitívne rekurzívnych funkčných symbolov.

(define (projection n x i)
  (if (= i n)
      (iter cdr (- n 1) x)
      (car (iter cdr (- i 1) x))))

(define (interpreter e x)
  (cond [(Zx? e) 0]
        [(Sx? e) (+ x 1)]
        [(Ix? e) (projection (Ix-arity e) x (Ix-i e))]
        [(Px? e) (cons (interpreter (Px-g e) x) (interpreter (Px-gs e) x))]
        [(Cx? e) (interpreter (Cx-h e) (interpreter (Cx-gs e) x))]
        [(Rx? e) (let ([y (car x)] [z (cdr x)])
                   (if (= y 0)
                       (interpreter (Rx-g e) z)
                       (interpreter (Rx-h e) (cons (- y 1) (cons (interpreter e (cons (- y 1) z)) z)))))]
        [(Xc2? e) (Xc2-m e)]
        [(Xpair? e) x]))

